//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <cxxtls/strtool.h>
#include <stdlib.h>
#include <stdio.h>

///@file
///
///  This file defines functions which are declared but not implemented
///  in the header, strtool.

using namespace std;

namespace cxxtls
{


string
StrTool::
expand_tabs(char const *s, bool *non_printable_chars, char replacement )
{
  string rv;

  int count=0;

  char c;

  while(0 != (c = *s++) )
  {
    if(c == '\t')
    {

      // int next_tab_stop = ( (count >> 3) + 1 ) << 3;

      // from expand.c:   next_tab_column = column + tab_size - column % tab_size;

      int next_tab_stop = count + 8 - count % 8;

      while(count != next_tab_stop)
      {
        rv += ' ';
        ++count;
      }

    }
    else
    {
      if(c < ' ' || c >= 0x7f )
      {
        if(non_printable_chars)
          *non_printable_chars = true;
        
        if(replacement)
          c = replacement;
      }

      rv += c;
      ++count;
    }
  }

  return rv;

}


string
StrTool::
pack_tabs(char const *s, bool stop_after_first_non_blank)
  // convert spaces to tabs if possible
{

  string rv;

  int output=0;  // output is always less than or equal to input
  int input=0;

  char space=0x20;

  char c;

  for(;;)  // process all characters in the string -- even the null
  {
    c = *s++;

    if(c == space)
    {
       ++input;  // allow output and input to differ by the number of
                 // blanks
    }
    else
    if(c == '\t')
    {
       input >>= 3;  // treat tabs as a bunch of input spaces
       ++input;
       input <<= 3;
    }
    else
    {
      // not a space: force sync between output and input character positions

      if( (input - output == 1) && 
          (output%8 == 7) 
        )
      {
        rv += space;
        ++output;
      }
      else
      {

        int tout = output >> 3;
        int tin  = input  >> 3;
        
        if(tout != tin)
        {
           while(tout != tin)
           {
             rv += '\t';   // the most recent output character is in a different
                           // tab stop region ( 8 char wide ) than the input
                           // character
             ++tout;
           }
          
           output = tout << 3;
          
        }
      }


      while(output != input)
      {
        rv += space;            // output and input are in same tab stop
        ++output;             // but some number of blanks apart
      }


      if(c == 0)
      {
        break;
      }

      rv += c;

      ++input;

      output = input;

      // now output = input and all blanks and tabs are accounted for.

      if(stop_after_first_non_blank)
      {
        rv += s;
        return rv;
      }

    }

  }

  return rv;
}

void
StrTool::
remove_leading(std::string &s, char space)
{
  // remove all leading copies of space from the string

  std::string::iterator scan = s.begin(),
                        end  = s.end();

  while(scan != end && *scan == space)
    ++scan;

  if(scan != s.begin())
    s.erase(s.begin(), scan);

}



bool 
StrTool::
snatch_file_info(std::string const &s, std::string &outputFile, int &outputLine)
   {
          // return true on success

          size_t rLength = s.size();

          if(rLength == 0)
            return false;

          size_t firstNonBlank = s.find_first_not_of(" \t");

          if(firstNonBlank > rLength)
            return false;
        
          std::string tmp(s.begin() + firstNonBlank, s.end());  
        
          tmp += ' ';  // make sure there's a blank at the end
        
          char name[3000];
          char numbuff[3000];
          int  number;
        
          name[0] = 0;
        
          //
          //  use scanf to parse some standard program output
          //  formats:
          //
          //      filename:lineNo:  -- grep format
          //      file lineNo:      -- borland c output
          //      "file", line 2:   -- other compiler outputs
          //      file(line)        -- microsoft
          //      #line 12 "file"   -- C preprocessor
          //      <file>            -- #include file names
          //      #12 "file"
          //      Error 112: "reports_pkg.c", line 48
          //      Error (112): "reports_pkg.c", line 48
          //      error: "reports_pkg.c", line 48
          //      warning: "reports_pkg.c", line 48
          //      file:             -- grep without the -n
          //      (file:line)       -- valgrind logs with debug on
          //      (in file)         -- valgrind logs without debug on
        
          if( 2 == sscanf(tmp.c_str(), "(%[^:]:%d)", name, &number) )
          {

            // valgrind log
          }
          else
          if( 1 == sscanf(tmp.c_str(), "(in %[^):]", name) )
          {
            number=1;
          }
          else
          if( 2 ==
              sscanf(tmp.c_str(), "error: \"%[+-.$#@!`~=a-zA-Z0-9_/\\]\", line %[0-9]",
                     name,
                     numbuff
                    )
            )
            {  //      error: "reports_pkg.c", line 48
        
               number=atol(numbuff);
            }
          else
          if( 2 ==
              sscanf(tmp.c_str(), "warning: \"%[+-.$#@!`~=a-zA-Z0-9_/\\]\", line %[0-9]",
                     name,
                     numbuff
                    )
            )
            {  // warning: "reports_pkg.c", line 48
        
               number=atol(numbuff);
            }
          else
          if( 2 ==
              sscanf(tmp.c_str(), "Error %*[()0-9]: \"%[+-.$#@!`~=a-zA-Z0-9_/\\]\", line %[0-9]",
                     name,
                     numbuff
                    )
            )
            {  // Error (112): "reports_pkg.c", line 48
               number=atol(numbuff);
            }
          else
          if( 2 ==
              sscanf(tmp.c_str(), "# %[0-9] \"%[^\"]\"",
                     numbuff,
                     name
                    )
            )
            { // #123 "file.c"
        
              // compiler -E output format
        
              number=atol(numbuff);
            }
          else
          if( 2 ==
              sscanf(tmp.c_str(), "#line %[0-9] \"%[^\"]\"",
                     numbuff,
                     name
                    )
            )
            { // #line 132 "file.c"
        
              // manually overrride of the compilers line number info
        
              number=atol(numbuff);
            }
          else
          if( 2 ==
              sscanf(tmp.c_str(), "\"%[+-.$#@!`~=a-zA-Z0-9_/\\]\", line %[0-9]",
                     name,
                     numbuff
                    )
            )
              { // "filename", line 47:
        
                 // aix compiler
        
                  number = atol(numbuff);
              }
          else
          if( 2 ==
              sscanf(tmp.c_str(), "%[^ :]:%d:", name, &number)
            )
              { // file:line:
        
                  // gnu grep output
        
              }
          else
          if( 2 == sscanf(tmp.c_str(), "%s %d", name, &number) )
              { // filename linenumber
        
                  // borland c++ compiler
        
              }
          else
          if( 2 == sscanf(tmp.c_str(), "%[^(](%d)", name, &number) )
              { // filename(linenumber)
        
                  // microsoft c++ compiler
        
              }
          else
          if( 1 == sscanf(tmp.c_str(), "\"%[^\"]\"", name) )
              { // "filename"
        
                  number=1;   // plain old quoted string
        
              }
          else
          if( 1 == sscanf(tmp.c_str(), "<%[^>]>", name) )
          {
            number=1;
          }
          else
          if( 1 == sscanf(tmp.c_str(), "%[^: ]:", name) )  // WARNING:  must be last real test!
              { // filename:
        
                  number=1;   // plain old filename ending in :
        
                  // fix erroneous return of true in some cases...
        
                  int l = strlen(name);
        
                  while(l)
                  {
                    --l;
                    if(name[l] != ' ')
                      break;
                  }
        
                  name[l+1] = 0;
        
#ifdef _MSC_VER
        
                  if(l == 0 && tmp[1] == ':' )
                {
                  if( (name[0] >= 'a' && name[0] <= 'z')
                      || (name[0] >= 'A' && name[0] <= 'Z')
                    )
                  {
                    // yep, we have a dos style fully qualified pathname:
                    //
                    //   c:\dir\file:
                    //
                    // And possibly this:
                    //
                    //  c:\dir\file:line
        
                    std::string drive = tmp.substr(0,2);
                    std::string path  = tmp.substr(2,tmp.size()-2);
        
                    if(2 == sscanf(path.c_str(), "%[^ :]:%d:", name, &number ) )
                    {
                      drive += name;
        
                      strcpy(name, drive.c_str());
        
                    }
                    else
                    {
                      drive = tmp.substr(0,tmp.size()-1);
        
                      strcpy(name, drive.c_str());
                    }
        
        
                  }
                }
        
#endif
        
        
              }
          else  // WARNING:  do not put any new tests here, put them before
                // the above test due to bugs in sscanf on some machines
              {
                // plain old non-quoted string
        
                  number = 1;
                  sscanf(tmp.c_str(), "%s", name);
                
              }

       outputFile = name;
       outputLine = number;
       
       return !outputFile.empty();


   }

bool 
StrTool::
getline(FILE *f, std::string &line)
{
     char buffer[32768];

     buffer[0] = 0;

     line.resize(0);

     bool someDataRead = false;

     for(;;)
     {
         char *p = fgets(buffer, sizeof(buffer), f);

         if(!p)
         {
             break;
         }

         someDataRead = true;

         size_t l = strlen(buffer);

         if(!l)
         {
            // if isatty(f->fileNo_), then sleep 1/10'th second

            continue;  // no read error, but no data?  Continue hoping to eventually get a newline
                       // this could happen on a terminal device but will burn up a lot of cpu.
         }
                      

         line += buffer;

     }

     if(!line.empty())
     {
        // remove the trailing \r\n

        if(line[line.size()-1] == '\n')
           line.erase(line.size()-1);

        #ifdef _MSC_VER

        if(!line.empty() && line[line.size()-1] == '\r')
           line.erase(line.size()-1);

        #endif

     }

     
     return someDataRead;
    
}

void 
StrTool::
fold(std::string const &input, size_t maxLen, std::vector<std::string> &output)
{
   size_t remainingLength = input.size();

   char const *scan = input.data();
   char const *end  = scan + remainingLength;

   if(remainingLength < maxLen)
   {
      output.push_back(input);   // no need to split
      return;
   }

   while(scan != end)
   {
       remainingLength = end - scan;

       if(remainingLength <= maxLen)
       {
          // remaining characters fit into one line -- push it out and quit

          if(remainingLength)
          {
              output.push_back(std::string());

              output.back().append(scan,end);
          }
          return;
       }
       else
       {
           //we must split

           output.push_back(std::string());

           size_t splitLength = maxLen;

           while(splitLength > 0)
           {
               char c = scan[splitLength];

               bool isWord = false;

               if(   (c >= 'A' && c <= 'Z')
                  || (c >= 'a' && c <= 'z')
                  || (c == '_')
                  || (c == ':')
                  || (c == ';')                  // for types and variables
                  || (c == '<')
                  || (c == '>')
                  || (c >= '0' && c <= '9')
                 )
               {
                  isWord = true;
               }

               if(!isWord)
                  break;

               --splitLength;

           }

           if(splitLength == 0)
           {
              splitLength = maxLen;  // try again with a reduced word character set


              while(splitLength > 0)
              {
                  char c = scan[splitLength];
              
                  bool isWord = false;
              
                  if(   (c >= 'A' && c <= 'Z')
                     || (c >= 'a' && c <= 'z')
                     || (c == '_')
                     || (c == ';')                  // for types and variables
                     || (c >= '0' && c <= '9')
                    )
                  {
                     isWord = true;
                  }
              
                  if(!isWord)
                     break;
              
                  --splitLength;
              
              }


              splitLength = maxLen;  // no word to split on -- just give up
           }

           output.back().append(scan, scan + splitLength);

           scan += splitLength;

       }

   }



}  // fold


void 
StrTool::
trim(std::string &s, char const *chars)
{
    if(s.empty())
       return;

    size_t firstNonMatching = s.find_first_not_of(chars);

    if(firstNonMatching > 0)
    {
       if(firstNonMatching < s.size())
       {
          s.erase(0, firstNonMatching); // remove leading blanks
       }
       else
       {
          // the string was all blanks

          s.resize(0);
          return;
       }
    }

    size_t lastNonMatching = s.find_last_not_of(chars);

    if(lastNonMatching >= s.size())
    {
       return;  // no trailing blanks
    }

    s.erase(lastNonMatching+1);

    return;

    
}



} // namespace cxxtls
